// CADWindow.cpp: implementation of the CCADWindow class.
//
//////////////////////////////////////////////////////////////////////

#include "CADWindow.h"
#include "resource.h"
#include <stdio.h>
#include <math.h>
#include <sg.h>

#pragma warning(disable : 4996) // sprintf

char LayerName[255];
double Scales[17] = {0.001, 0.0025, 0.005, 0.01, 0.02, 0.03, 0.05, 0.1, 0.25, 0.4, 
	1.0, 2.5, 5.0, 10.0, 25.0, 50.0, 100.0};

BOOL CALLBACK LayerNameDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) 
{ 
	switch (message) 
	{
		case WM_INITDIALOG:       
			SetDlgItemText(hwnd, IDC_EDIT1, "BOARD");       			
			return TRUE; 

		case WM_CLOSE:             
			EndDialog(hwnd, TRUE);             
			return TRUE; 

		case WM_COMMAND:             
			if(LOWORD((DWORD)wParam) == IDOK)             
			{                 
				GetDlgItemText(hwnd, IDC_EDIT1, LayerName, 255);  
				EndDialog(hwnd, TRUE);                 
				return TRUE;             
			} 
			if(LOWORD((DWORD)wParam) == IDCANCEL)             
			{                 			
				EndDialog(hwnd, FALSE);                 
				return TRUE;             
			} 
		break;     
	}      
	return FALSE;      
} 


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CADENUM CADEnum;
CADCREATE CADCreate;
CADCLOSE CADClose;
CADIS3D CADIs3D;
CADGETLASTERROR CADGetLastError;
CADPROHIBITCURESASPOLY CADProhibitCurvesAsPoly;
CADLAYOUTCURRENT CADLayoutCurrent;
CADLAYOUTNAME CADLayoutName;
CADLAYOUTCOUNT CADLayoutCount;
CADGETBOX CADGetBox;
CADSETSHXOPTIONS CADSetSHXOptions;

CCADWindow::CCADWindow(LPCTSTR szClassName, WNDPROC WndProc, HINSTANCE hInst,
                 HICON hIcon, HICON hCursor, LPCTSTR lpszMenuName,
                 HBRUSH color, UINT style):
CWindow(szClassName, WndProc, hInst, hIcon, hCursor, lpszMenuName, color, style)
{
	ScaleID = ID_FITWINDOW;
	offset.x = 2;
	offset.y = TOOLBAR_HEIGHT;
	PredPoint.x = 0;
	PredPoint.y = 0;
	DrwMode = 0;
	CircleDrwMode = 0;	
	Drag = FALSE;
	Cursor = LoadCursorA(WndClass.hInstance, "CURSOR1");
	HInst = hInst;
	ScaleRect.x = -1;
	nIsInsideInsert = 0;
	hwndToolBar = 0;
	hwndComboBox = 0;
	hwndLayoutsLabel = 0;
	hwndStatusBar = 0;
	bDoubleBuffered = true;
	bGetArcsAsCurves = false;
	bGetTextsAsCurves = false;

	hCAD = 0;
	hCADImporter = LoadLibrary("CADImporter.dll");
	if (hCADImporter)
	{
		CADEnum = (CADENUM) GetProcAddress(hCADImporter, "CADEnum");
		CADCreate = (CADCREATE) GetProcAddress(hCADImporter, "CADCreate");
		CADClose = (CADCLOSE) GetProcAddress(hCADImporter, "CADClose");	
		CADIs3D = (CADIS3D) GetProcAddress(hCADImporter, "CADIs3D");
		CADGetLastError = (CADGETLASTERROR) GetProcAddress(hCADImporter, "CADGetLastError");
		CADProhibitCurvesAsPoly = (CADPROHIBITCURESASPOLY) GetProcAddress(hCADImporter, "CADProhibitCurvesAsPoly");
		CADLayoutCount = (CADLAYOUTCOUNT) GetProcAddress(hCADImporter, "CADLayoutCount");
		CADLayoutCurrent = (CADLAYOUTCURRENT) GetProcAddress(hCADImporter, "CADLayoutCurrent");
		CADLayoutName = (CADLAYOUTNAME) GetProcAddress(hCADImporter, "CADLayoutName");
		CADGetBox = (CADGETBOX) GetProcAddress(hCADImporter, "CADGetBox");
		CADSetSHXOptions = (CADSETSHXOPTIONS) GetProcAddress(hCADImporter, "CADSetSHXOptions");
	}
	else
		MessageBoxA(0, (LPCTSTR)"CADImporter.dll not loaded!", "Error", 0);	
	
}

CCADWindow::~CCADWindow()
{
  if (hCADImporter)
	  FreeLibrary(hCADImporter);
}

void CCADWindow::DoCreateToolBar(HWND hwndParent, HINSTANCE hInst) 
{	
	LPCTSTR tbrCaption = "Layouts:";

	if (!hwndToolBar)
	{
		hwndToolBar = ::CreateWindowEx(
			WS_EX_WINDOWEDGE,      
			TOOLBARCLASSNAME,
			(LPCTSTR) NULL,  
			CCS_TOP | WS_CHILD | WS_BORDER | WS_CAPTION, 
			0, 0, 0, TOOLBAR_HEIGHT,        
			hwndParent,    
			NULL,
			hInst,         
			NULL);         

		hwndLayoutsLabel = ::CreateWindowEx(
			0,      
			"STATIC",
			tbrCaption,  
			WS_CHILD,			 
			4, 4, LABELLAYOUTWIDTH, TOOLBAR_BUTTON_SIZE,
			hwndToolBar,    
			NULL,
			hInst,         
			NULL);        

		::ShowWindow(hwndToolBar, SW_SHOWNORMAL);    
		::ShowWindow(hwndLayoutsLabel, SW_SHOWNORMAL); 
	}
}

void CCADWindow::DoCreateComboBox(HINSTANCE hInst) 
{	
	if (!hwndToolBar) return;
	if (hCAD) 
	{
		char LayoutName[100];
		int i, currLayout, Count;		
		Count = CADLayoutCount(hCAD);
		if (hwndComboBox) 
			DestroyWindow(hwndComboBox);		
		hwndComboBox = ::CreateWindowEx( 
			0,      
			"COMBOBOX",
			(LPCTSTR) NULL,  
			WS_CHILD | CBS_DROPDOWNLIST,			 
			LABELLAYOUTWIDTH + 4, 4, 200, 40+COMBOBOX_ITEM_SIZE*Count,        
			hwndToolBar,    
			NULL,
			hInst,         
			NULL);
	    for (i= 0; i < Count; ++i)
	    {
			CADLayoutName(hCAD, i, LayoutName, 100);			
			::SendMessage(hwndComboBox, (UINT) CB_ADDSTRING, 0, (LPARAM) (LPCTSTR) LayoutName);			
		}
		CADLayoutCurrent(hCAD, (int *)&currLayout, false);
		::SendMessage(hwndComboBox, (UINT) CB_SETCURSEL, (WPARAM) currLayout, 0);
		RePaint();
		::ShowWindow(hwndComboBox, SW_SHOWNORMAL);
	}	
}

void CCADWindow::DoCreateStatusBar(HWND hwndParent, HINSTANCE hInst)
{     
	if (!hwndStatusBar)
		hwndStatusBar = ::CreateWindowEx( 
			0,                        
			STATUSCLASSNAME,          
			(LPCTSTR) NULL,           
			WS_CHILD,
			0, 0, 0, 0,               
			hwndParent,               
			NULL,                     
			hInst,                    
			NULL);                    

	SplitStatusBar(); 
	SendMessage(hwndStatusBar, SB_SETTEXT, 0, (LPARAM)NULL);
	SendMessage(hwndStatusBar, SB_SETTEXT, 1, (LPARAM)NULL);
	SendMessage(hwndStatusBar, SB_SETTEXT, 2, (LPARAM)"Demo"); 
	ShowWindow(hwndStatusBar, SW_SHOWNORMAL);
}

void CCADWindow::SplitStatusBar(int nParts)
{
	RECT rect; 
    HLOCAL hloc; 
    LPINT lpParts; 
    int i, nWidth; 
	
    GetClientRect(hWnd, &rect); 
    hloc = LocalAlloc(LHND, sizeof(int) * nParts); 
    lpParts = (LPINT)LocalLock(hloc);  

    nWidth = rect.right / nParts; 
    for (i = 0; i < nParts; i++) 
	{ 
        lpParts[i] = nWidth; 
        nWidth+= nWidth; 
    } 	
	
    ::SendMessage(hwndStatusBar, SB_SETPARTS, (WPARAM) nParts, (LPARAM) lpParts); 	

    LocalUnlock(hloc); 
    LocalFree(hloc);   
}

bool CCADWindow::SetCurrentLayout()
{
    int lindex = ::SendMessage(hwndComboBox, (UINT) CB_GETCURSEL, 0, 0);
    CADLayoutCurrent(hCAD, (int *)&lindex, true);
	RecalculateExtents();
	RePaint();
	return true;
}

void CCADWindow::Load()
{
	TCHAR FileName[100];
	FileName[0] = 0;
	char cErrorCode[ERR_MSG_LEN];
	char str[60];
	int nErrorCode;

	if (this->GetFile (this->hWnd, FileName, (LPCTSTR)"CAD files (dwg, dxf, rtl, spl, prn, gl2, hpgl2, hpgl, hp2, hp1, hp, plo, hpg, hg, hgl, plt, cgm, svg)\0*.dxf;*.dwg;*.rtl;*.spl;*.prn;*.gl2;*.hpgl2;*.hpgl;*.hp2;*.hp1;*.hp;*.plo;*.hpg;*.hg;*.hgl;*.plt;*.cgm;*.svg\0") == 0) return;
	if (hCAD) CADClose(hCAD);

	if (hCADImporter)
	{
		SendMessage(hwndStatusBar, SB_SETTEXT, 0, (LPARAM)"Please wait...");
		CADSetSHXOptions("","","", 1,1);// please call it before CADCreate
		hCAD = CADCreate(hWnd, FileName);
		nErrorCode = CADGetLastError(cErrorCode);
		if (nErrorCode == 0)
		{
			CADProhibitCurvesAsPoly(hCAD, (int(bGetArcsAsCurves)+1)%2);// 1 => permit conversion of arcs to polyline			
			ChangeScale(ID_FITWINDOW);
			DoCreateToolBar(hWnd, HInst);
			DoCreateComboBox(HInst);
			RecalculateExtents();
			sprintf(str, "(%6.6f x %6.6f)\0", fAbsWidth, fAbsHeight);
			SendMessage(hwndStatusBar, SB_SETTEXT, 0, (LPARAM)FileName);
			SendMessage(hwndStatusBar, SB_SETTEXT, 1, (LPARAM)str);
		}
		else
		{
			SendMessage(hwndStatusBar, SB_SETTEXT, 0, (LPARAM)cErrorCode);
			switch(nErrorCode)
			{			
			case 1006:
				MessageBox(hWnd, cErrorCode, "Error", MB_ICONERROR);
				SendMessage(hwndStatusBar, SB_SETTEXT, 0, (LPARAM)"No file loaded");
			default:
				break;
			}
		}
		RePaint();
	}
}

void CCADWindow::ReSize(WPARAM wParam,LPARAM lParam)
{
	SplitStatusBar();
  	::SendMessage(hwndStatusBar, WM_SIZE, wParam, lParam);
	::SendMessage(hwndToolBar, TB_AUTOSIZE, (WPARAM)0 , (LPARAM)0 );
}

void CCADWindow::Draw(HWND Window, HDC DC)
{
RECT wndrect;
PARAM s;
PAINTSTRUCT paint;

	if (hCAD)
	{
		s.hWnd = Window;
    s.hDC = DC;
    if (!DC) s.hDC = BeginPaint(Window, &paint);
		s.offset = this->offset;
		s.Scale = Scale;
		s.DrawMode = DrwMode;
		s.CircleDrawMode = CircleDrwMode;
		s.GetArcsCurves = bGetArcsAsCurves;
		s.GetTextsCurves = bGetTextsAsCurves;
		s.IsInsideInsert = &nIsInsideInsert;
		GetClientRect(hWnd, &wndrect);
		FillRect(s.hDC, &wndrect, WndClass.hbrBackground);
		SetMapMode(s.hDC, MM_ANISOTROPIC);
		SetViewportOrgEx(s.hDC, 0, 0, 0);
		CADEnum(hCAD, (int(bGetTextsAsCurves) << 3), DoDraw, &s);
    if (!DC) EndPaint(Window, &paint);
	}
}

void CCADWindow::BeginDrag(POINTS Point)
{
	if (hCAD) 
	{
		SetCapture(hWnd);
		Drag = true;
		PredPoint = Point;
		SetCursor(Cursor);
	}
}

void CCADWindow::EndDrag(POINTS Point)
{
	if (hCAD) 
	{
		ReleaseCapture();
		Drag = false;
		SetCursor(WndClass.hCursor);
	}
}

void CCADWindow::MouseMove(POINTS Point)
{
	if (Drag)
	{
		SetCursor(Cursor);
		offset.x = offset.x + Point.x - PredPoint.x;
		offset.y = offset.y + Point.y - PredPoint.y;
		PredPoint = Point;
		RePaint();
	}
}

void CCADWindow::ChangeScale(LONG ID)
{
	HMENU hMenu;
	hMenu = GetMenu(hWnd);
	CheckMenuItem(hMenu, ScaleID + ID_SCALE + 1, MF_BYCOMMAND | MF_UNCHECKED);
	ScaleID = ID - ID_SCALE - 1;
	CheckMenuItem(hMenu, ScaleID + ID_SCALE + 1, MF_BYCOMMAND | MF_CHECKED);
	if (ID == ID_FITWINDOW) 
		FitToWindow();
	else
	{
		Scale = Scales[ScaleID];
		RecalculateExtents();
		RePaint();
	}
}

void CCADWindow::RecalculateExtents()
{
	PredPoint.x = 0;
	PredPoint.y = 0; //TOOLBAR_HEIGHT;
	offset.x = 0;
	offset.y = TOOLBAR_HEIGHT;
	CADGetBox(hCAD, &BoxLeft, &BoxRight, &BoxTop, &BoxBottom);
	fAbsWidth = BoxRight - BoxLeft;
	fAbsHeight = BoxTop - BoxBottom;
	fCoef = fAbsHeight / fAbsWidth;
	ScaleRect.x = -1;
}

void CCADWindow::FitToWindow()
{
	double L, T, R, B, W, H;
	RECT rect;
	if (CADGetBox(hCAD, &L, &R, &T, &B) == 0) return;
    W = R - L;
    if (W == 0) W = 1;
	H = T - B;
    if (H == 0) H = 1;
	GetClientRect(hWnd, &rect);
    W = abs(rect.right - rect.left) / W;
    H = abs(rect.bottom - rect.top) / H;
    if (W > H) W = H;
	RecalculateExtents();
	Scale = W;	
	RePaint();
}

void CCADWindow::GetLayers()
{    
	if (hCAD)
	{			
		if (DialogBox(HInst, 
			MAKEINTRESOURCE(IDD_DIALOG1), 
			hWnd, 
			(DLGPROC)LayerNameDlgProc) == 0) return; 		
		LAYER L;
		L.Name = (char *)malloc(strlen(LayerName) + 1);
		strcpy(L.Name, LayerName);		
		L.Count = 0;	  
		CADEnum(hCAD, 0, GetLayerFromData, &L);	  	  
		char  buffer[200];     
		sprintf(buffer, "There are %d entities on %s layer", L.Count, L.Name);   	  
		MessageBox(hWnd, buffer, "", 0);	  
		free(L.Name);
	}
}

void CCADWindow::DrawingMode(LONG mode)
{

	HMENU hMenu = GetMenu(hWnd);
	CheckMenuItem(hMenu, ID_DRAWINGSTYLE_DASHDOTS, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_DRAWINGSTYLE_WINAPI, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_DRAWINGSTYLE_SOLIDLINES, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, mode, MF_BYCOMMAND | MF_CHECKED);
	switch(mode)
		{
			case ID_DRAWINGSTYLE_DASHDOTS:				
				DrwMode = 0;
				break;
			case ID_DRAWINGSTYLE_WINAPI:
				DrwMode = 1;
				break;
			case ID_DRAWINGSTYLE_SOLIDLINES:
				DrwMode = 2;
				break;
		}			
	RePaint();
}
void CCADWindow::Circle(LONG mode)
{
	HMENU hMenu = GetMenu(hWnd);
	CheckMenuItem(hMenu, ID_CIRCLE_DRAWASARC, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_CIRCLE_DRAWASPOLY_DASHDOTS, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_CIRCLE_DRAWASPOLY_WINAPI, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_CIRCLE_DRAWASPOLY_SOLIDLINES, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, mode, MF_BYCOMMAND | MF_CHECKED);
	switch(mode)
		{
			case ID_CIRCLE_DRAWASARC:				
				CircleDrwMode = 1;
				CheckMenuItem(hMenu, ID_CIRCLE_DRAWASPOLY_WINAPI, MF_BYCOMMAND | MF_CHECKED);
				break;
			case ID_CIRCLE_DRAWASPOLY_DASHDOTS:
				CircleDrwMode = 0;
				break;
			case ID_CIRCLE_DRAWASPOLY_WINAPI:
				CircleDrwMode = 1;
				CheckMenuItem(hMenu, ID_CIRCLE_DRAWASARC, MF_BYCOMMAND | MF_CHECKED);
				break;
			case ID_CIRCLE_DRAWASPOLY_SOLIDLINES:				
				CircleDrwMode = 2;
				break;
		}			
	RePaint();
}

void CCADWindow::ArcsAsCurves(LONG mode)
{
	HMENU hMenu = GetMenu(hWnd);
	CheckMenuItem(hMenu, ID_GETARCSASCURVES, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_GETARCSASPOLY, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, mode, MF_BYCOMMAND | MF_CHECKED);

	switch (mode)
	{
	case ID_GETARCSASCURVES:
		bGetArcsAsCurves = true;
		CADProhibitCurvesAsPoly(hCAD, 0);
		break;
	case ID_GETARCSASPOLY:
		bGetArcsAsCurves = false;
		CADProhibitCurvesAsPoly(hCAD, 1);
		break;
	}
	RePaint();
}

void CCADWindow::TextsAsCurves()
{
	HMENU hMenu = GetMenu(hWnd);
    if (bGetTextsAsCurves)
       CheckMenuItem(hMenu, ID_GETTEXTSASCURVES, MF_BYCOMMAND | MF_UNCHECKED);
    else
       CheckMenuItem(hMenu, ID_GETTEXTSASCURVES, MF_BYCOMMAND | MF_CHECKED);
	bGetTextsAsCurves = !bGetTextsAsCurves;
	RePaint();
}

void CCADWindow::SwitchDblBuf()
{
  if (bDoubleBuffered)
    CheckMenuItem(GetMenu(hWnd), ID_VIEW_DBLBUF, MF_BYCOMMAND | MF_UNCHECKED);
  else
    CheckMenuItem(GetMenu(hWnd), ID_VIEW_DBLBUF, MF_BYCOMMAND | MF_CHECKED);
  bDoubleBuffered = !bDoubleBuffered;
}

int CCADWindow::WMPaint(HWND wnd, int Msg, int wParam, int lParam)
{
	HDC DC, MemDC;
	HBITMAP MemBitmap;
	HGLOBAL OldBitmap;
	PAINTSTRUCT PS;
	RECT Rect;

	if ((!bDoubleBuffered) || (wParam)) //wParam }=> HDC
	{
		Draw(wnd, (HDC)wParam);
		return 0;
	}
	else
	{
		DC = GetDC(0);
		GetWindowRect(wnd, &Rect);
		InflateRect(&Rect, -Rect.left, -Rect.top);
		MemBitmap = CreateCompatibleBitmap(DC, Rect.right, Rect.bottom);
		ReleaseDC(0, DC);
		MemDC = CreateCompatibleDC(0);
		OldBitmap = SelectObject(MemDC, MemBitmap);
		DC = BeginPaint(wnd, &PS);
		DefWindowProcA(wnd, WM_ERASEBKGND, (unsigned int)MemDC, (unsigned int)MemDC);
		SendMessage(wnd, Msg, (int)MemDC, (int)lParam);
		BitBlt(DC, 0, 0, Rect.right, Rect.bottom, MemDC, 0, 0, SRCCOPY);
		EndPaint(wnd, &PS);    
		SelectObject(MemDC, OldBitmap);
		DeleteDC(MemDC);
		DeleteObject(MemBitmap);
		return 0;
	}
}
